import logging
import os
import shutil

import deltatar.deltatar as deltatar
import deltatar.crypto   as crypto
import deltatar.tarfile  as tarfile

from . import BaseTest

def flip_bits (fname, off, b=0x01, n=1):
    """
    Open file *fname* at offset *off*, replacing the next *n* bytes with
    their values xor’ed with *b*.
    """
    fd = os.open (fname, os.O_RDWR)

    try:
        pos = os.lseek (fd, off, os.SEEK_SET)
        assert pos == off
        chunk = os.read (fd, n)
        chunk = bytes (map (lambda v: v ^ b, chunk))
        os.write (fd, chunk)
    finally:
        os.close (fd)


def gz_header_size (fname, off=0):
    """
    Determine the length of the gzip header starting at *off* in file fname.

    The header is variable length because it may contain the filename as NUL
    terminated bytes.
    """
    # length so we need to determine where the actual payload starts
    off = tarfile.GZ_HEADER_SIZE
    fd = os.open (fname, os.O_RDONLY)

    try:
        pos = os.lseek (fd, off, os.SEEK_SET)
        assert pos == off
        while os.read (fd, 1)[0] != 0:
            off += 1
            pos = os.lseek (fd, off, os.SEEK_SET)
            assert pos == off
    finally:
        os.close (fd)

    return off


class RecoverTest (BaseTest):
    """
    Disaster recovery: restore corrupt backups.
    """

    GIT_DIR = '.git'

    def setUp(self):
        '''
        Create base test data
        '''
        self.pwd = os.getcwd()
        os.system('rm -rf target_dir source_dir* backup_dir* huge')
        os.makedirs('source_dir/test/test2')
        self.hash = dict()
        self.hash["source_dir/test/test2"] = ''
        self.hash["source_dir/big"]  = self.create_file("source_dir/big", 50000)
        self.hash["source_dir/small"]  = self.create_file("source_dir/small", 100)
        self.hash["source_dir/test/huge"]  = self.create_file("source_dir/test/huge", 700000)
        self.hash["source_dir/test/huge2"]  = self.create_file("source_dir/test/huge2", 800000)

        self.consoleLogger = logging.StreamHandler()
        self.consoleLogger.setLevel(logging.DEBUG)

        if not os.path.isdir(self.GIT_DIR):
            # Not running inside git tree, take our
            # own testing directory as source.
            self.GIT_DIR = 'testing'
            if not os.path.isdir(self.GIT_DIR):
                raise Exception('No input directory found: ' + self.GIT_DIR)


    def test_recover_corrupt_byte (self):
        """
        Flip a bit in a non-header byte of a backup set, then recover.

        Expects the extraction to fail in normal mode. With disaster recovery,
        extraction must succeed, and exactly one file must be missing.
        """
        mode           = "#gz"
        dst_path       = "source_dir"
        src_path       = "%s2" % dst_path
        bak_path       = "backup_dir"
        backup_file    = "the_full_backup.tar.gz"
        backup_full    = "%s/%s" % (bak_path, backup_file)
        index_file     = "the_full_index.gz"

        def vname (*a, **kwa):
            return backup_file

        dtar = deltatar.DeltaTar (mode="#gz",
                                  logger=self.consoleLogger,
                                  index_name_func=lambda _: index_file,
                                  volume_name_func=vname)
        self.hash = dict ()
        os.makedirs (src_path)
        for i in range (5):
            f = "dummy_%rd" % i
            self.hash [f] = self.create_file ("%s/%s" % (src_path, f), 5 + i)

        dtar.create_full_backup \
            (source_path=src_path, backup_path=bak_path)

        # first restore must succeed
        dtar.restore_backup(target_path=dst_path,
                            backup_tar_path=backup_full)
        for key, value in self.hash.items ():
            f = "%s/%s" % (dst_path, key)
            assert os.path.exists (f)
            assert value == self.md5sum (f)
        shutil.rmtree (dst_path)
        shutil.rmtree (src_path)

        flip_bits (backup_full, gz_header_size (backup_full) + 1)

        # normal restore must fail
        curdir = os.getcwd () # not restored after below failure
        with self.assertRaises (Exception) as cm:
            dtar.restore_backup(target_path=dst_path,
                                backup_tar_path=backup_full)
        assert type (cm.exception) in [ tarfile.CompressionError
                                      , tarfile.ReadError
                                      ]

        os.chdir (curdir)
        # but recover will succeed
        failed = dtar.recover_backup(target_path=dst_path,
                                     backup_indexes_paths=[
                                         "%s/%s" % (bak_path, index_file)
                                     ])
        assert len (failed) == 1

        # with one file missing
        missing = []
        for key, value in self.hash.items ():
            kkey = "%s/%s" % (dst_path, key)
            if os.path.exists (kkey):
                assert value == self.md5sum (kkey)
            else:
                missing.append (key)
        assert len (missing) == 1

        shutil.rmtree (dst_path)

